Require Import ssreflect ssrfun ssrnat ssrbool eqtype mtac.
Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

Structure dynamic := D { ty : Type; el : ty }.

Structure result A B x t := R { fu : A -> B; pf : t = fu x }.
Implicit Arguments R [A B x t].

Set Printing Existential Instances.

Lemma abs_app 
  (A : Type) (x : A) (A' : Type) r (t1 : A' -> ty r) (t2 : A') (r1 : result x t1) (r2 : result x t2):
  result x (t1 t2).
Proof.
elim r1. intros f1 p1.
elim r2. intros f2 p2.
rewrite p1 p2.
exact (R (fun y=>f1 y (f2 y)) (Logic.eq_refl _)).
Defined.

Lemma match_eq :
   forall A B : Type,
   forall x : A,
   forall (r : dynamic) (b : bool) (P Q : ty r),
   result x b ->
   result x P ->
   result x Q ->
   result x (if b then P else Q).
Proof.
intros A B x r b P Q r1 r2 r3.
elim r1; intros f1 b1.
elim r2; intros f2 b2.
elim r3; intros f3 b3.
rewrite b1 b2 b3.
exact (R (fun y=>if (f1 y) then (f2 y) else f3 y) (Logic.eq_refl _)).
Defined.

Implicit Arguments match_eq [A x r b P Q].

(** Abstracts term [x] from [t], returning a function (fun y=>t[y/x]) *)
Definition abstract A B (x : A) (t : B) :=
  let: f :=
    mfix loop [ r : dynamic ] : T (result x (el r)) :=
    mmatch r as r' return T (result x (el r')) with
    | D x => 
      ret (R id (Logic.eq_refl _))
    | [A' (t1 : A' -> ty r) t2] D (t1 t2)  =>
        r1 <- loop (D t1);
        r2 <- loop (D t2);
        ret (abs_app r1 r2)
    | [b (P:ty r) (Q:ty r)] D (match b with
          | true => P
          | false => Q
          end)
      => 
      b' <- loop (D b);
      P' <- loop (D P);
      Q' <- loop (D Q);
      ret (match_eq B b' P' Q') 
    | [z] z => 
      ret (R (fun _=>el z) (Logic.eq_refl _))
    end in
  f (D t).
  
Lemma eq_fu (A : Type) (x y : A) (P : Prop) (r : result x P) :
  x = y -> fu r y -> P.
Proof. by elim: r=>f /= -> ->. Qed.

(** Rewrite Mtactic. Changes goal [P] with goal [Q] by rewrite [x] with [y]. *)
Definition rewrite A (x y : A) (p : T (x = y)) (P Q : Prop) (q : Q) : T P :=
  r <- abstract x P;
  p' <- p;
  a <- mmatch Q as Q' return T ((Q':Prop) -> P) with
       | fu r y => ret (eq_fu (r:=r) p')
       end;
  ret (a q). 

Implicit Arguments rewrite [A x y P Q q].



(** Example of a (dumb) tactic to rewrite nat equalities. *)

Program
Definition eq_nats  :=
  mfix f [x : nat; y : nat] : T (x == y) :=
  r <-
    mmatch (x, y) as p return T ((x, y) = p -> x == y) with
    | (x, x) => ret (fun H=>_)
    | [x1 x2] (x1 + x2, x2 + x1) => 
      ret (fun H =>_)
    end;
  ret (r (erefl _)).
Next Obligation.
  by move=> _ x y [->]; apply: eq_refl.
Qed.
Next Obligation.
  move=> _ x y x1 x2 [-> ->].
  apply/eqP.
  by apply: addnC.
Qed.

Example plus_S n m : (n + m == m + n) && (m == m) && (n == n). 
Proof.
  apply: (run (rewrite (eq_nats _ _)))=>/=.
  apply: (run (rewrite (eq_nats _ _)))=>/=.
  apply: (run (rewrite (eq_nats _ _)))=>/=.
  by [].
Qed.
